/*
 * vector3.h
 * begin: 01.09.2005
 *
 * copyright (C) 2005 by Jan Ehrenberger < lankvil@seznam.cz >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

/*
 * C00rs
 * prepracovane coors. matematicky spravne vektory a souradnice
 */

#ifndef C00R_VECTOR3_H
#define C00R_VECTOR3_H

#include <cmath>
#include "coor3.h"

namespace c00r
{
	template <class G>
	class coor3;

	/** vector3
	 * vektory pro 3D
	 */
	template <class G>
	class vector3 {
	public:
		G x, y, z;

		/* constructors: */
		vector3 (const G &a, const G &b, const G &c)
			:x(a), y(b), z(c) 
		{ }
		vector3 (const vector3<G> &a)
			:x(a.x), y(a.y), z(a.z)
		{ }
		/* rozdil dvou souradnic je vektor */
		vector3 (const coor3<G> &a, const coor3<G> &b)
			:x(a.x - b.x), y(a.y - b.y), z(a.z - b.z)
		{ }
		vector3 ()
			:x(0), y(0), z(0)
		{ }

		G magnitude () const
		{ return (G)std::sqrt(x*x + y*y + z*z); }
		void set_magnitude (G nm)
		{
			float m = magnitude();
			float a = (m) ? (nm/m) : 0;
			x *= a;
			y *= a;
			z *= a;
		}
		void normalize ()
		{
			float m = magnitude();
			if (m == 0) return;
			x /= m;
			y /= m;
			z /= m;
		}

		/* rotace kolem osy y - v radianech */
		void roty(G angle)
		{
			G x0 = x,
			  b = (G)cosf(angle),
			  a = (G)sinf(angle);

			x0  = b * x0 + a * z;
			z  = b * z - a * x0;
		}
		
		/* rotace kolem libovolne osy - v radianech */
		void rot (const G& angle, const vector3<G>& a)
    		{
			G x1 = x,
			  y1 = y, 
			  z1 = z;

			// Calculate the sine and cosine of the angle once
			G cosTheta = (float)cosf(angle);
			G sinTheta = (float)sinf(angle);

			// Find the new x position for the new rotated point
			x1  = (cosTheta + (1 - cosTheta) * a.x*a.x) * x;
			x1 += ((1 - cosTheta) * a.x * a.y - a.z * sinTheta) * y;
			x1 += ((1 - cosTheta) * a.x * a.z + a.y * sinTheta) * z;

			// Find the new y position for the new rotated point
			y1  = ((1 - cosTheta) * a.x * a.y + a.z * sinTheta) * x;
			y1 += (cosTheta + (1 - cosTheta) * a.y*a.y) * y;
			y1 += ((1 - cosTheta) * a.y * a.z -a.x * sinTheta) * z;

			// Find the new z position for the new rotated point
			z1  = ((1 - cosTheta) * a.x * a.z - a.y * sinTheta) * x;
			z1 += ((1 - cosTheta) * a.y * a.z + a.x * sinTheta) * y;
			z1 += (cosTheta + (1 - cosTheta) * a.z * a.z) * z;

			x = x1;
			y = y1;
			z = z1;
		} 

		operator bool () const
		{ return x || y || z; }
		vector3 operator- ()
		{ return vector3<G>(-x, -y, -z); }

		vector3& operator += (const vector3<G> &a)
		{ x += a.x; y += a.y; z += a.z; return *this; }
		vector3& operator -= (const vector3<G> &a)
		{ x -= a.x; y -= a.y; z -= a.z; return *this; }

	};

	template <class G>
	bool operator == (const vector3<G> &a, const vector3<G> &b)
	{ return (a.x == b.x) && (a.y == b.y) && (a.z == a.y); }

	template <class G>
	bool operator != (const vector3<G> &a, const vector3<G> &b)
	{ return (a.x != b.x) || (a.y != b.y) || (a.z != a.y); }

	/* scitani vektoru */
	template <class G>
	vector3<G> operator + (const vector3<G> &a, const vector3<G> &b)
	{ return vector3<G>(a.x+b.x, a.y+b.y, a.z+b.z); }
	/* odcitani vektoru */
	template <class G>
	vector3<G> operator - (const vector3<G> &a, const vector3<G> &b)
	{ return vector3<G>(a.x-b.x, a.y-b.y, a.z-b.z); }

	/* nasobeni vektoru cislem */
	template <class G>
	vector3<G> operator * (const vector3<G> &a, const G &b)
	{ return vector3<G>(a.x*b, a.y*b, a.z*b); }
	/* deleni vektoru cislem */
	template <class G>
	vector3<G> operator / (const vector3<G> &a, const G &b)
	{ return vector3<G>(a.x/b, a.y/b, a.z/b); }

	/* vektorovy soucin */
	template <class G>
	vector3<G> x (const vector3<G> &a, const vector3<G> &b)
	{
		return vector3<G>(
			a.y * b.z - a.z * b.y,
			a.z * b.x - a.x * b.z,
			a.x * b.y - a.y * b.x
		);
	}


} // namespace c00r

#endif // C00R_VECTOR3_H


