/*
 * mymem.h
 * begin: 26.7.2003
 *
 * copyright (C) 2003-2004 by Jan Ehrenberger < Honza_E@atlas.cz >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#ifndef H4E3R_MYMEM_H
#define H4E3R_MYMEM_H

#include <sys/types.h>
#include <exception>

namespace h4e3r
{


  /*
   * ENullPointer
   * - vyjimka naznacuje, ze doslo k pokusu o zpristupneni dat na nullovem
   *   pointeru
   */
  class ENullPointer :public std::exception
  {
    static char *_comment;

  public:
    ENullPointer() throw() { }
    ~ENullPointer() throw() { }
    
    const char *what() const throw()
      { return _comment; }
  };

  /*
   * handle_base
   * - zaklad vsech handlu pameti
   */
  class handle_base
  {
  protected:
    /*
     * CWrapper
     * - struktura uchovavajici alokovanou pamet a pocet ukazatelu na ni
     */
    class CWrapper
    {
      int m_counter;
    public:
      CWrapper() :m_counter(0) { }
      virtual ~CWrapper() { }
      virtual void* get_ptr() = 0;
      void incr() { m_counter++; }                // increment counter
      bool decr() { return (bool)(--m_counter); } // decrement counter
    };

    CWrapper* rep;

    void _free()
      { if( rep && !rep->decr() ) delete rep; }
    void _alloc(CWrapper* w)
      { _free(); rep=w; rep->incr(); }
    void test_rep() const throw(ENullPointer)
      { if( !rep ) throw ENullPointer(); }

  public:
    handle_base() :rep(0) { }
    handle_base(void* p) :rep((CWrapper*)p) { if(rep) rep->incr(); }
    handle_base(CWrapper*& r) :rep(r) { if(rep) rep->incr(); }
    handle_base(const handle_base& a)
      :rep(a.rep)
      { if(rep) rep->incr(); }
    ~handle_base() { _free(); }

    void free() { _free(); rep = 0; }
    void* make_void() { return (void*)rep; }
    bool isNull()
      { return !(bool)rep; }

    handle_base& operator=(handle_base a);
    bool operator==(const handle_base& m) const { return (rep==m.rep); }
    operator bool() const { return (bool)rep; }
  friend void convert(handle_base&, const handle_base&);
  friend void* make_void(const handle_base& m)
    { return (m.rep)? (void*)m.rep : 0; }
  };


  /*
   * mem_handle
   * - ukazatel na volnou pamet
   * - zajistuje alokovani a odstraneni pameti
   */
  template <class __type>
  class mem_handle :public handle_base
  {
    class CWrapper_typal :public CWrapper
    {
    protected:
      __type* m_rep;
    public:
      CWrapper_typal() :m_rep(new __type) { }
      CWrapper_typal(__type* ptr) :m_rep(ptr) { } // predany ukazatel musi byt vytvoreny pomoci operatoru new!
      ~CWrapper_typal() { delete m_rep; }

      void* get_ptr() { return (void*)m_rep; }
    };

  public:
    mem_handle()
      :handle_base((CWrapper*)(new CWrapper_typal)) { }
    mem_handle(bool a) // a==true - alokuje pamet; a==false - nealokuje pamet
      { if(a) handle_base::_alloc(new CWrapper_typal); }
    mem_handle(__type* ptr) // predany ukazatel musi byt vytvoreny pomoci opertaoru new!
      :handle_base((CWrapper*)(new CWrapper_typal(ptr))) { }
    mem_handle(const mem_handle& a)
      :handle_base(a) { }
    ~mem_handle() { }

    __type* get_ptr() { return (rep) ? (__type*)rep->get_ptr() : 0; }
    const __type* get_ptr() const { return (rep) ? (__type*)rep->get_ptr() : 0; }
    __type& get_ref() { test_rep(); return *(__type*)rep->get_ptr(); }
    const __type& get_ref() const { test_rep(); return *(__type*)rep->get_ptr(); }

    void alloc()
      { _alloc(new CWrapper_typal); }

    __type* operator->() { test_rep(); return (__type*)rep->get_ptr(); }
    const __type* operator->() const { test_rep(); return (__type*)rep->get_ptr(); }
    __type& operator*() { test_rep(); return *(__type*)rep->get_ptr(); }
    const __type& operator*() const { test_rep(); return *(__type*)rep->get_ptr(); }
  };


  /*
   * array_handle
   * - ukazatel na pole ve volne pameti
   * - zajistuje alokovani a odstraneni pameti
   */
  template <class __type>
  class array_handle :public handle_base
  {
    class CWrapper_typal :public CWrapper
    {
    protected:
      __type* m_rep;
      size_t m_size;
    public:
      CWrapper_typal(const size_t& n) :m_rep(new __type[n]), m_size(n) { }
      ~CWrapper_typal() { delete[] m_rep; }

      void* get_ptr() { return (void*)m_rep; }
      const size_t& size() const { return m_size; }
    };
  public:
    array_handle()
      :handle_base() { }
    array_handle(const size_t& n)
      :handle_base(new CWrapper_typal(n))  { }
    array_handle(const array_handle& a)
      :handle_base(a) { }
    ~array_handle() { }

    __type* get_ptr() { return (rep) ? (__type*)rep->get_ptr() : 0; }
    const __type* get_ptr() const { return (rep) ? (__type*)rep->get_ptr() : 0; }
    size_t size() const { return (rep) ? ((CWrapper_typal*)rep)->size() : 0; }

    void alloc(int n)
      { _alloc(new CWrapper_typal(n)); }

    __type& operator[](const size_t& i) { test_rep(); return ((__type*)rep->get_ptr())[i]; }
    const __type& operator[](const size_t& i) const { test_rep(); return ((__type*)rep->get_ptr())[i]; }
  };


  /* uprava array_handle<char> pro lepsi zpracovani retezcu */
  class string_handle :public array_handle<char>
  {
  public:
    string_handle()
      :array_handle<char>() { }
    string_handle(const string_handle& a)
      :array_handle<char>(a) { }
    string_handle(const char *);
    string_handle(const char *, const size_t& len);
    ~string_handle() { }

    const char* s() const /* dela to samy jako get_ptr const, ale zapis je o neco kratsi */
      { return get_ptr(); }
    const char* cs() const
      { return get_ptr(); }
    char* s() /* dela to samy jako get_ptr, ale zapis je o neco kratsi */
      { return get_ptr(); }
    size_t len() const; /* vrati pocet znaku v retezci :) */

    const string_handle& copy(const char *);
    const string_handle& copy(const char *, const size_t& len);
    const string_handle& copy(const string_handle& a) /* 'hluboka' kopie retezce */
      { return copy(a.get_ptr()); }
    /* porovnani ve stylu C tak, jak ho zname
     * kdyz je parametr mensi, vraci cislo mensi nez nula, kdyz je stejny, vraci nulu
     * a kdyz je vetsi, vraci cislo vetsi nez nula */
    int cmp(const char *) const;
    int cmp(const string_handle& a) const
      { return cmp(a.get_ptr()); }
    /* shoda retezcu
     * pro porovnavani retezcu v zadnem pripade nepouzivejte ==, protoze to porovnava
     * pouze ukazatele a ne obsah retezcu. radeji pouzijte nasledujici funkci */
    bool equal(const string_handle& a) const
      { return (cmp(a) == 0); }
    bool equal(const char* a) const
      { return (cmp(a) == 0); }

    /* porovnavaci operatory */
    bool operator>(const string_handle& a) const
      { return (cmp(a)<0); }
    bool operator>=(const string_handle& a) const
      { return (cmp(a)<=0); }
    bool operator<(const string_handle& a) const
      { return (cmp(a)>0); }
    bool operator<=(const string_handle& a) const
      { return (cmp(a)>=0); }
    /* ty same porovnavaci operatory pro Ceckove retezce */
    bool operator>(const char *a) const
      { return (cmp(a)<0); }
    bool operator>=(const char* a) const
      { return (cmp(a)<=0); }
    bool operator<(const char* a) const
      { return (cmp(a)>0); }
    bool operator<=(const char* a) const
      { return (cmp(a)>=0); }
  };

/*
  template <class __t>
  mem_handle<__t> make_from_void(void* p)
  {
    mem_handle<__t> m;
    convert(m, handle_base(p));
    return m;
  }

  template <class __h>
  __h __make_from_void(void* p)
  {
    __h m;
    convert(m, handle_base(p));
    return m;
  }
*/

} /* namespace h4e3r */

namespace mymem = h4e3r;


#endif /* H4E3R_MYMEM_H */
